@@ -27,6 +27,8 @@ doBench (name, KEM p) = bgroup name
     , bench "encapsulate (batch)" $ perBatchEnv (const setupEncap) (return . runEncap)
     , bench "decapsulate" $ perRunEnv setupDecap (return . runDecap)
     , bench "decapsulate (batch)" $ perBatchEnv (const setupDecap) (return . runDecap)
+    , env (generate p) $ bench "encoding (public)" . nf (encodeDecode p) . fst
+    , env (generate p) $ bench "encoding (private)" . nf (encodeDecode p) . snd
     ]
   where
     gen32 = getRandomBytes 32 :: IO Bytes
@@ -46,6 +48,13 @@ doBench (name, KEM p) = bgroup name
         (_, c) <- encapsulate ek
         return (dk, c)
 
+    encodeDecode :: (Encode obj, Decode obj, Eq (obj a), ParamSet a)
+                 => proxy a -> obj a -> Bool
+    encodeDecode prx k = decode prx (aligned $ encode k) == Just k
+
+    aligned :: Bytes -> Bytes
+    aligned = id
+
 main :: IO ()
 main = defaultMain
     [ bgroup "mlkem" $ map doBench kems
