@@ -1,3 +1,10 @@
+-- |
+-- Module      : Auxiliary
+-- License     : BSD-3-Clause
+-- Copyright   : (c) 2025 Olivier Chéron
+--
+-- ML-KEM auxiliary functions
+--
 {-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE DataKinds #-}
@@ -24,7 +31,6 @@ import Crypto.Hash.Algorithms
 
 import Data.ByteArray (ByteArrayAccess, Bytes, View)
 import qualified Data.ByteArray as B
-import qualified Data.Memory.Endian as B
 
 import Control.DeepSeq (NFData(..))
 import Control.Monad
@@ -39,10 +45,13 @@ import GHC.TypeNats
 import Foreign.Ptr (Ptr, plusPtr)
 import Foreign.Storable (pokeByteOff)
 
+import Unsafe.Coerce
+
 import Block (blockIndex)
 import BlockN (BlockN, MutableBlockN)
 import Builder (Builder)
 import Crypto (BlockDigest)
+import Machine
 import Marking (Classified, SecurityMarking(..), Leak(..))
 import SecureBlock (SecureBlock)
 import SecureBytes (SecureBytes)
@@ -169,9 +178,11 @@ infixl 6 ..+, ..-
 -- adding noise to secret information.
 (..+) :: Rq Sec -> Rq Sec -> Rq Pub
 Rq a ..+ Rq b = Rq $ BlockN.zipWith (.+) a b
+{-# NOINLINE (..+) #-}
 
 (..-) :: Rq Pub -> Rq Sec -> Rq Sec
 Rq a ..- Rq b = Rq $ BlockN.zipWith (.-) a b
+{-# NOINLINE (..-) #-}
 
 instance Leak Rq
 
@@ -198,14 +209,16 @@ instance Classified marking => Add (Tq marking) where
 
 instance Leak Tq
 
-instance Classified marking => NFData (Tq marking) where
+instance NFData (Tq marking) where
     rnf (Tq a) = toNormalForm a
 
 instance BiMul (Tq Pub) (Tq Sec) where
     (..*) = multiplyNTTs
+    {-# NOINLINE (..*) #-}
 
 instance BiMulAdd (Tq Pub) (Tq Sec) where
     biMulAdd = multiplyNTTsAdd
+    {-# NOINLINE biMulAdd #-}
 
 #ifdef ML_KEM_TESTING
 instance Mul (Tq Sec) where
@@ -367,14 +380,14 @@ decompress d y = Zq $ fromIntegral (x2d `unsafeShiftR` d)
 -- Compress a polynomial with 𝑑 < 12
 rcompress :: Classified marking => Int -> Rq marking -> BlockN marking N Word16
 rcompress !d (Rq a) = BlockN.map (compress d) a
-{-# SPECIALIZE rcompress :: Int -> Rq Sec -> BlockN Sec N Word16 #-}
-{-# SPECIALIZE rcompress :: Int -> Rq Pub -> BlockN Pub N Word16 #-}
+{-# SPECIALIZE NOINLINE rcompress :: Int -> Rq Sec -> BlockN Sec N Word16 #-}
+{-# SPECIALIZE NOINLINE rcompress :: Int -> Rq Pub -> BlockN Pub N Word16 #-}
 
 -- Decompress a polynomial with 𝑑 < 12
 rdecompress :: Classified marking => Int -> BlockN marking N Word16 -> Rq marking
 rdecompress !d = Rq . BlockN.map (decompress d)
-{-# SPECIALIZE rdecompress :: Int -> BlockN Sec N Word16 -> Rq Sec #-}
-{-# SPECIALIZE rdecompress :: Int -> BlockN Pub N Word16 -> Rq Pub #-}
+{-# SPECIALIZE NOINLINE rdecompress :: Int -> BlockN Sec N Word16 -> Rq Sec #-}
+{-# SPECIALIZE NOINLINE rdecompress :: Int -> BlockN Pub N Word16 -> Rq Pub #-}
 
 -- Generates a pseudorandom element of T𝑞 from a seed and two indices
 sampleNTT :: SecureBytes Pub -> Word8 -> Word8 -> Tq Pub
@@ -383,7 +396,7 @@ sampleNTT seed !x !y = runST $ do
     runXof b (280 * 3) 0 0
     Tq <$> BlockN.unsafeFreeze b
   where
-    runXof !b !xofLen !pos !j = case someNatVal (8 * fromIntegral xofLen) of
+    runXof !b !xofLen !pos !j = case someNatVal (fromIntegral (8 * xofLen)) of
         SomeNat proxy -> do
             let bytes = Crypto.unBlockDigest (doHash proxy)
             loop b xofLen bytes pos j
@@ -414,22 +427,14 @@ sampleNTT seed !x !y = runST $ do
         pokeByteOff d (len + 1) y
     len = B.length seed
 
-wordBits :: Int
-wordBits = finiteBitSize (0 :: Word)
+peekWord :: Ptr WordLE -> ST s WordM
+peekWord p = fromLE <$> ST.peek p
 
-wordBytes :: Int
-wordBytes = div wordBits 8
+peekWordPos :: Ptr WordLE -> BitPos -> ST s WordM
+peekWordPos a bp = fromLE <$> ST.peekElemOff a (wordOff bp)
 
-type WordLE = B.LE Word
-
-peekWord :: Ptr WordLE -> ST s Word
-peekWord p = B.unLE <$> ST.peek p
-
-peekWordPos :: Ptr WordLE -> BitPos -> ST s Word
-peekWordPos a bp = B.unLE <$> ST.peekElemOff a (wordOff bp)
-
-pokeWordPos :: Ptr WordLE -> BitPos -> Word -> ST s ()
-pokeWordPos a bp = ST.pokeElemOff a (wordOff bp) . B.LE
+pokeWordPos :: Ptr WordLE -> BitPos -> WordM -> ST s ()
+pokeWordPos a bp = ST.pokeElemOff a (wordOff bp) . toLE
 
 newtype BitPos = BitPos Int
 
@@ -450,10 +455,12 @@ nextPos :: Int -> BitPos -> (Int, BitPos)
 nextPos requested (BitPos p) = (howMany, BitPos $ p + howMany)
   where howMany = availPos requested (BitPos p)
 
-getMask :: Int -> Word
+getMask :: Int -> WordM
 getMask howMany
     | howMany >= wordBits = maxBound
     | otherwise = (1 `unsafeShiftL` howMany) - 1
+    -- branch useful only when processing one byte at a time due to
+    -- architecture not supporting unaligned memory access
 
 -- Takes a seed as input and outputs a pseudorandom sample from the
 -- distribution D_eta
@@ -480,15 +487,19 @@ samplePolyCBD !eta !input = runST $ ST.withByteArray input $ \p -> do
             getBits p bp' (acc + fromIntegral (popCount bits)) (j - howMany)
 
 -- Encodes an array of 𝑑-bit integers into a byte array for 1 ≤ 𝑑 ≤ 12
-byteEncode :: Classified marking => Int -> BlockN marking N Word16 -> Builder marking
-byteEncode !d !f = Builder.create (32 * d) $ \b ->
-    outer b zeroPos 0 0
+byteEncode :: Int -> BlockN marking N Word16 -> Builder marking
+byteEncode d f = Builder.create (32 * d) (runByteEncode d f)
+{-# INLINE byteEncode #-}
+
+runByteEncode :: Int -> BlockN marking N Word16 -> Ptr WordLE -> ST s ()
+runByteEncode !d !f dst = outer dst zeroPos 0 0
   where
-    outer :: Ptr WordLE -> BitPos -> Word -> Int -> ST s ()
+    outer :: Ptr WordLE -> BitPos -> WordM -> Int -> ST s ()
     outer !b !bp !o pos = when (pos < n) $
         inner b pos bp o (BlockN.index f (Offset pos)) d
+    {-# NOINLINE outer #-}
 
-    inner :: Ptr WordLE -> Int -> BitPos -> Word -> Word16 -> Int -> ST s ()
+    inner :: Ptr WordLE -> Int -> BitPos -> WordM -> Word16 -> Int -> ST s ()
     inner !b !pos !bp !o !a j
         | j == 0 = outer b bp o (pos + 1)
         | bitPos bp + howMany < wordBits = inner b pos bp' o' a' j'
@@ -499,15 +510,16 @@ byteEncode !d !f = Builder.create (32 * d) $ \b ->
         o' = o .|. (x `unsafeShiftL` bitPos bp)
         a' = a `unsafeShiftR` howMany
         j' = j - howMany
-{-# SPECIALIZE byteEncode :: Int -> BlockN Sec N Word16 -> Builder Sec #-}
-{-# SPECIALIZE byteEncode :: Int -> BlockN Pub N Word16 -> Builder Pub #-}
 
 -- Optimization of byteEncode when 𝑑=1
 byteEncode1 :: BlockN Sec N Word16 -> Builder Sec
-byteEncode1 !f = Builder.create 32 $ \b ->
-    loop b 0 0
+byteEncode1 !f = Builder.create 32 (runByteEncode1 f)
+{-# INLINE byteEncode1 #-}
+
+runByteEncode1 :: BlockN marking N Word16 -> Ptr WordLE -> ST s ()
+runByteEncode1 !f dst = loop dst 0 0
   where
-    loop :: Ptr WordLE -> Word -> Int -> ST s ()
+    loop :: Ptr WordLE -> WordM -> Int -> ST s ()
     loop !b !o pos
         | pos == n = return ()
         | bitPos bp + 1 < wordBits = loop b o' (pos + 1)
@@ -519,11 +531,15 @@ byteEncode1 !f = Builder.create 32 $ \b ->
         a = BlockN.index f (Offset pos)
 
 -- byteEncode with 𝑑=12 after conversion from the field
-byteEncode12 :: Classified marking => Tq marking -> Builder marking
-byteEncode12 (Tq f) = byteEncode 12 $ BlockN.map (\(Zq x) -> x) f
+byteEncode12 :: Tq marking -> Builder marking
+byteEncode12 = byteEncode 12 . fromField
+  where
+    fromField :: Tq marking -> BlockN marking N Word16
+    fromField (Tq f) = unsafeCoerce f
+{-# INLINE byteEncode12 #-}
 
 -- Decodes a byte array into an array of 𝑑-bit integers for 1 ≤ 𝑑 ≤ 12
-byteDecode :: forall marking ba. Classified marking => ByteArrayAccess ba => Int -> ba -> BlockN marking N Word16
+byteDecode :: forall marking ba. (Classified marking, ByteArrayAccess ba) => Int -> ba -> BlockN marking N Word16
 byteDecode !d !b = runST $
     ST.withByteArray b $ \p -> do
         f <- BlockN.new (Proxy :: Proxy marking)
@@ -543,12 +559,12 @@ byteDecode !d !b = runST $
                 j' = j + howMany
             inner f p i bp' v' j'
 
-    get :: Ptr WordLE -> BitPos -> Int -> ST s Word
+    get :: Ptr WordLE -> BitPos -> Int -> ST s WordM
     get p bp howMany = do
         x <- (`unsafeShiftR` bitPos bp) <$> peekWordPos p bp
         return (x .&. getMask howMany)
-{-# SPECIALIZE byteDecode :: forall ba. ByteArrayAccess ba => Int -> ba -> BlockN Sec N Word16 #-}
-{-# SPECIALIZE byteDecode :: forall ba. ByteArrayAccess ba => Int -> ba -> BlockN Pub N Word16 #-}
+{-# SPECIALIZE byteDecode :: forall ba. ByteArrayAccess ba => Int -> View ba -> BlockN Sec N Word16 #-}
+{-# SPECIALIZE byteDecode :: forall ba. ByteArrayAccess ba => Int -> View ba -> BlockN Pub N Word16 #-}
 {-# SPECIALIZE byteDecode :: Int -> View Bytes -> BlockN Sec N Word16 #-}
 {-# SPECIALIZE byteDecode :: Int -> View Bytes -> BlockN Pub N Word16 #-}
 
@@ -565,7 +581,7 @@ byteDecode1 !b = runST $
         x <- peekWord p
         inner f (p `plusPtr` wordBytes) x i 0
 
-    inner :: MutableBlockN Sec N Word16 s -> Ptr WordLE -> Word -> Int -> Int -> ST s ()
+    inner :: MutableBlockN Sec N Word16 s -> Ptr WordLE -> WordM -> Int -> Int -> ST s ()
     inner !f !p !acc !i j
         | j == wordBits = outer f p i
         | otherwise = do
@@ -574,5 +590,7 @@ byteDecode1 !b = runST $
             inner f p (acc `unsafeShiftR` 1) (i + 1) (j + 1)
 
 -- byteDecode with 𝑑=12 and conversion to the field
-byteDecode12 :: Classified marking => ByteArrayAccess ba => ba -> Tq marking
+byteDecode12 :: (Classified marking, ByteArrayAccess ba) => ba -> Tq marking
 byteDecode12 = Tq . BlockN.map toZq . byteDecode 12
+{-# SPECIALIZE byteDecode12 :: ByteArrayAccess ba => View ba -> Tq Sec #-}
+{-# SPECIALIZE byteDecode12 :: ByteArrayAccess ba => View ba -> Tq Pub #-}
