@@ -1,3 +1,10 @@
+-- |
+-- Module      : ByteArrayST
+-- License     : BSD-3-Clause
+-- Copyright   : (c) 2025 Olivier Chéron
+--
+-- Byte array primitives in the @ST@ monad instead of @IO@
+--
 {-# LANGUAGE RankNTypes #-}
 module ByteArrayST
     ( unsafeCreate, withByteArray
@@ -18,8 +25,8 @@ unsafeCreate :: ByteArray ba => Int -> (forall s. Ptr p -> ST s ()) -> ba
 unsafeCreate sz f = B.unsafeCreate sz (stToIO . f)
 {-# INLINE unsafeCreate #-}
 
-withByteArray :: ByteArrayAccess ba => ba -> (forall s. Ptr p -> ST s a) -> ST t a
-withByteArray b f = unsafeIOToST $ B.withByteArray b (stToIO . f)
+withByteArray :: ByteArrayAccess ba => ba -> (Ptr p -> ST s a) -> ST s a
+withByteArray b f = unsafeIOToST $ B.withByteArray b (unsafeSTToIO . f)
 {-# INLINE withByteArray #-}
 
 peek :: Storable a => Ptr a -> ST s a
