@@ -1,10 +1,17 @@
+-- |
+-- Module      : SecureBlock
+-- License     : BSD-3-Clause
+-- Copyright   : (c) 2025 Olivier Chéron
+--
+-- Use a type-level annotation to decide between a scrubbed block or a regular
+-- block
+--
 {-# LANGUAGE CPP #-}
 module SecureBlock
-    ( SecureBlock, create, index, iterModify, Marking.map, new, thaw
+    ( SecureBlock, create, index, iterModify, new, thaw
     , unsafeCast, unsafeFreeze, toNormalForm
 #ifdef ML_KEM_TESTING
-    , eq, Marking.showsPrec, fromList, Marking.replicate, toList
-    , SecureBlock.length
+    , eq, Marking.showsPrec, toList, SecureBlock.length
 #endif
     ) where
 
