@@ -1,3 +1,12 @@
+-- |
+-- Module      : Main
+-- License     : BSD-3-Clause
+-- Copyright   : (c) 2025 Olivier Chéron
+--
+-- The ML-KEM test suite.  Can be instanciated twice, with and without the
+-- @ML_KEM_TESTING@ macro to run property testing with assertions enabled in
+-- the internal modules.
+--
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE DataKinds #-}
 {-# LANGUAGE ExistentialQuantification #-}
@@ -25,6 +34,8 @@ import Data.Word
 
 import GHC.TypeNats
 
+import Foreign.Ptr (plusPtr)
+
 import Auxiliary
 import BlockN (BlockN)
 import Builder (Builder)
@@ -112,15 +123,38 @@ byteDecodeBytes = byteDecode
 byteEncodeBytes :: Int -> BlockN Sec 256 Word16 -> Bytes
 byteEncodeBytes d = runBytes . byteEncode d
 
-byteEncodeBytes1 :: BlockN Sec 256 Word16 -> Bytes
-byteEncodeBytes1 = runBytes . byteEncode1
-
-byteEncodeBytes12 :: Tq Sec -> Bytes
-byteEncodeBytes12 = runBytes . byteEncode12
-
 runBytes :: Builder Sec -> Bytes
 runBytes = Builder.run . leak
 
+runUnaligned :: Builder Sec -> UnalignedBytes
+runUnaligned = Builder.runRelaxed . leak
+
+newtype UnalignedBytes = UnalignedBytes (B.View Bytes)
+    deriving (Eq,Ord,Show)
+
+instance Semigroup UnalignedBytes where
+    UnalignedBytes a <> UnalignedBytes b =
+        B.allocAndFreeze (na + nb) $ \p -> do
+            B.copyByteArrayToPtr a p
+            B.copyByteArrayToPtr b (p `plusPtr` na)
+      where
+        na = B.length a
+        nb = B.length b
+
+instance Monoid UnalignedBytes where
+    mempty = B.convert (B.empty :: Bytes)
+
+instance B.ByteArrayAccess UnalignedBytes where
+    length (UnalignedBytes v) = B.length v
+    withByteArray (UnalignedBytes v) = B.withByteArray v
+
+instance B.ByteArray UnalignedBytes where
+    allocRet n f = do
+        let build ba = UnalignedBytes (B.dropView ba offset)
+        (a, ba) <- B.allocRet (n + offset) $ \p -> f (p `plusPtr` offset)
+        return (a, build ba)
+      where offset = 1
+
 #endif
 
 data P = forall a. (ParamSet a, Show a) => P (Proxy a)
@@ -278,12 +312,20 @@ main = defaultMain $ testGroup "mlkem"
         , testGroup "encoding"
             [ testProperty "byteEncode . byteDecode == id" $ \(D d) -> do
                 b <- arbitraryBytes (32 * d)
-                return (b === byteEncodeBytes d (byteDecode d b))
+                return (b === runBytes (byteEncode d (byteDecode d b)))
+            , testProperty "byteEncode . byteDecode == id (unaligned)" $ \(D d) -> do
+                b <- arbitraryBytes (32 * d)
+                return (B.convert b === runUnaligned (byteEncode d (byteDecode d b)))
             , testProperty "byteEncode1 . byteDecode1 == id" $ do
                 b <- arbitraryBytes 32
-                return (b === byteEncodeBytes1 (byteDecode1 b))
+                return (b === runBytes (byteEncode1 (byteDecode1 b)))
+            , testProperty "byteEncode1 . byteDecode1 == id (unaligned)" $ do
+                b <- arbitraryBytes 32
+                return (B.convert b === runUnaligned (byteEncode1 (byteDecode1 b)))
             , testProperty "byteDecode12 . byteEncode12 == id" $ \(PolyNTT p) ->
-                p === byteDecode12 (byteEncodeBytes12 p)
+                p === byteDecode12 (runBytes $ byteEncode12 p)
+            , testProperty "byteDecode12 . byteEncode12 == id (unaligned)" $ \(PolyNTT p) ->
+                p === byteDecode12 (runUnaligned $ byteEncode12 p)
             , testProperty "byteEncode 8" $ \x ->
                 B.replicate 256 x === byteEncodeBytes 8 (BlockN.replicate $ fromIntegral x)
             , testCase "byteEncode 1 (zeros)" $
